@@ -25,6 +25,7 @@
 #include "nls.h"
 #include "blkdev.h"
 #include "common.h"
+#include "mbsalign.h"
 #include "fdisk.h"
 #include "wholedisk.h"
 
@@ -203,7 +204,7 @@ int	fd,				/* the disk */
 	ext_index,			/* the prime extended partition */
 	listing = 0,			/* no aborts for fdisk -l */
 	nowarn = 0,			/* no warnings for fdisk -l/-s */
-	dos_compatible_flag = ~0,
+	dos_compatible_flag = 0,	/* disabled by default */
 	dos_changed = 0,
 	partitions = 4;			/* maximum partition + 1 */
 
@@ -219,7 +220,7 @@ unsigned int	heads,
 	sector_factor = 1,
 	user_set_sector_size = 0,
 	units_per_sector = 1,
-	display_in_cyl_units = 1;
+	display_in_cyl_units = 0;
 
 unsigned long long total_number_of_sectors;	/* (!) 512-byte sectors */
 unsigned long grain = DEFAULT_SECTOR_SIZE,
@@ -258,14 +259,14 @@ void fatal(enum failure why) {
 " fdisk [options] -l <disk> list partition table(s)\n"
 " fdisk -s <partition>      give partition size(s) in blocks\n"
 "\nOptions:\n"
-" -b <size>                 sector size (512, 1024, 2048 or 4096)\n"
-" -c                        switch off DOS-compatible mode\n"
-" -h                        print help\n"
-" -u <size>                 give sizes in sectors instead of cylinders\n"
-" -v                        print version\n"
-" -C <number>               specify the number of cylinders\n"
-" -H <number>               specify the number of heads\n"
-" -S <number>               specify the number of sectors per track\n"
+" -b <size>             sector size (512, 1024, 2048 or 4096)\n"
+" -c[=<mode>]           compatible mode: 'dos' or 'nondos' (default)\n"
+" -h                    print this help text\n"
+" -u[=<unit>]           display units: 'cylinders' or 'sectors' (default)\n"
+" -v                    print program version\n"
+" -C <number>           specify the number of cylinders\n"
+" -H <number>           specify the number of heads\n"
+" -S <number>           specify the number of sectors per track\n"
 "\n");
 			break;
 		case unable_to_open:
@@ -568,9 +569,19 @@ void list_types(struct systypes *sys)
 	i = done = 0;
 
 	do {
-		printf("%c%2x  %-15.15s", i ? ' ' : '\n',
-		        sys[next].type, _(sys[next].name));
- 		next = last[i++] + done;
+		#define NAME_WIDTH 15
+		char name[NAME_WIDTH * MB_LEN_MAX];
+		size_t width = NAME_WIDTH;
+
+		printf("%c%2x  ", i ? ' ' : '\n', sys[next].type);
+		size_t ret = mbsalign(_(sys[next].name), name, sizeof(name),
+				      &width, MBS_ALIGN_LEFT, 0);
+		if (ret == (size_t)-1 || ret >= sizeof(name))
+			printf("%-15.15s", _(sys[next].name));
+		else
+			fputs(name, stdout);
+
+		next = last[i++] + done;
 		if (i > 3 || next >= last[i]) {
 			i = 0;
 			next = ++done;
@@ -753,7 +764,7 @@ void update_units(void)
 	if (display_in_cyl_units && cyl_units)
 		units_per_sector = cyl_units;
 	else
-		units_per_sector = 1; 	/* in sectors */
+		units_per_sector = 1;	/* in sectors */
 }
 
 static void
@@ -785,18 +796,12 @@ warn_alignment(void) {
 "the physical sector size. Aligning to a physical sector (or optimal\n"
 "I/O) size boundary is recommended, or performance may be impacted.\n"));
 
-	if (dos_compatible_flag) {
+	if (dos_compatible_flag)
 		fprintf(stderr, _("\n"
 "WARNING: DOS-compatible mode is deprecated. It's strongly recommended to\n"
-"         switch off the mode (command 'c')"));
+"         switch off the mode (with command 'c')."));
 
-		if (display_in_cyl_units)
-			fprintf(stderr, _(" and change display units to\n"
-"         sectors (command 'u').\n"));
-		else
-			fprintf(stderr, ".\n");
-
-	 } else if (display_in_cyl_units)
+	if (display_in_cyl_units)
 		fprintf(stderr, _("\n"
 "WARNING: cylinders as display units are deprecated. Use command 'u' to\n"
 "         change units to sectors.\n"));
@@ -965,8 +970,6 @@ get_topology(int fd) {
 #ifdef HAVE_LIBBLKID_INTERNAL
 	blkid_probe pr;
 
-	if (user_set_sector_size)
-		return;
 	pr = blkid_new_probe();
 	if (pr && blkid_probe_set_device(pr, fd, 0, 0) == 0) {
 		blkid_topology tp = blkid_probe_get_topology(pr);
@@ -995,12 +998,20 @@ get_topology(int fd) {
 #endif
 
 	if (user_set_sector_size)
-		;
-	else if (blkdev_get_sector_size(fd, &arg) == 0)
+		/* fdisk since 2.17 differentiate between logical and physical
+		 * sectors size. For backward compatibility the
+		 *    fdisk -b <sectorsize>
+		 * changes both, logical and physical sector size.
+		 */
+		phy_sector_size = sector_size;
+
+	else if (blkdev_get_sector_size(fd, &arg) == 0) {
 		sector_size = arg;
 
-	if (!phy_sector_size)
-		phy_sector_size = sector_size;
+		if (!phy_sector_size)
+			phy_sector_size = sector_size;
+	}
+
 	if (!min_io_size)
 		min_io_size = phy_sector_size;
 	if (!io_size)
@@ -1505,11 +1516,11 @@ read_int(unsigned int low, unsigned int dflt, unsigned int high,
 
 
 int
-get_partition(int warn, int max) {
+get_partition_dflt(int warn, int max, int dflt) {
 	struct pte *pe;
 	int i;
 
-	i = read_int(1, 0, max, 0, _("Partition number")) - 1;
+	i = read_int(1, dflt, max, 0, _("Partition number")) - 1;
 	pe = &ptes[i];
 
 	if (warn) {
@@ -1526,6 +1537,11 @@ get_partition(int warn, int max) {
 	return i;
 }
 
+int
+get_partition(int warn, int max) {
+	return get_partition_dflt(warn, max, 0);
+}
+
 static int
 get_existing_partition(int warn, int max) {
 	int pno = -1;
@@ -1556,14 +1572,17 @@ static int
 get_nonexisting_partition(int warn, int max) {
 	int pno = -1;
 	int i;
+	int dflt = 0;
 
 	for (i = 0; i < max; i++) {
 		struct pte *pe = &ptes[i];
 		struct partition *p = pe->part_table;
 
 		if (p && is_cleared_partition(p)) {
-			if (pno >= 0)
+			if (pno >= 0) {
+				dflt = pno + 1;
 				goto not_unique;
+			}
 			pno = i;
 		}
 	}
@@ -1575,7 +1594,7 @@ get_nonexisting_partition(int warn, int max) {
 	return -1;
 
  not_unique:
-	return get_partition(warn, max);
+	return get_partition_dflt(warn, max, dflt);
 }
 
 const char *
@@ -1590,8 +1609,11 @@ void change_units(void)
 {
 	display_in_cyl_units = !display_in_cyl_units;
 	update_units();
-	printf(_("Changing display/entry units to %s\n"),
-		str_units(PLURAL));
+
+	if (display_in_cyl_units)
+		printf(_("Changing display/entry units to cylinders (DEPRECATED!)\n"));
+	else
+		printf(_("Changing display/entry units to sectors\n"));
 }
 
 static void
@@ -1611,7 +1633,7 @@ static void
 toggle_dos_compatibility_flag(void) {
 	dos_compatible_flag = ~dos_compatible_flag;
 	if (dos_compatible_flag)
-		printf(_("DOS Compatibility flag is set\n"));
+		printf(_("DOS Compatibility flag is set (DEPRECATED!)\n"));
 	else
 		printf(_("DOS Compatibility flag is not set\n"));
 
@@ -2637,7 +2659,8 @@ static void
 move_begin(int i) {
 	struct pte *pe = &ptes[i];
 	struct partition *p = pe->part_table;
-	unsigned int new, first;
+	unsigned int new, free_start, curr_start, last;
+	int x;
 
 	if (warn_geometry())
 		return;
@@ -2645,13 +2668,37 @@ move_begin(int i) {
 		printf(_("Partition %d has no data area\n"), i + 1);
 		return;
 	}
-	first = get_partition_start(pe);
-	new = read_int(first, first, first + get_nr_sects(p) - 1, first,
+
+	/* the default start is at the second sector of the disk or at the
+	 * second sector of the extended partition
+	 */
+	free_start = pe->offset ? pe->offset + 1 : 1;
+
+	curr_start = get_partition_start(pe);
+
+	/* look for a free space before the current start of the partition */
+	for (x = 0; x < partitions; x++) {
+		unsigned int end;
+		struct pte *prev_pe = &ptes[x];
+		struct partition *prev_p = prev_pe->part_table;
+
+		if (!prev_p)
+			continue;
+		end = get_partition_start(prev_pe) + get_nr_sects(prev_p);
+
+		if (!is_cleared_partition(prev_p) &&
+		    end > free_start && end <= curr_start)
+			free_start = end;
+	}
+
+	last = get_partition_start(pe) + get_nr_sects(p) - 1;
+
+	new = read_int(free_start, curr_start, last, free_start,
 		       _("New beginning of data")) - pe->offset;
 
 	if (new != get_nr_sects(p)) {
-		first = get_nr_sects(p) + get_start_sect(p) - new;
-		set_nr_sects(p, first);
+		unsigned int sects = get_nr_sects(p) + get_start_sect(p) - new;
+		set_nr_sects(p, sects);
 		set_start_sect(p, new);
 		pe->changed = 1;
 	}
@@ -2875,7 +2922,7 @@ main(int argc, char **argv) {
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
-	while ((c = getopt(argc, argv, "b:cC:hH:lsS:uvV")) != -1) {
+	while ((c = getopt(argc, argv, "b:c::C:hH:lsS:u::vV")) != -1) {
 		switch (c) {
 		case 'b':
 			/* Ugly: this sector size is really per device,
@@ -2893,7 +2940,12 @@ main(int argc, char **argv) {
 			user_cylinders = atoi(optarg);
 			break;
 		case 'c':
-			dos_compatible_flag = 0;
+			dos_compatible_flag = 0;	/* default */
+
+			if (optarg && !strcmp(optarg, "=dos"))
+				dos_compatible_flag = ~0;
+			else if (optarg && strcmp(optarg, "=nondos"))
+				fatal(usage);
 			break;
 		case 'h':
 			fatal(help);
@@ -2915,7 +2967,11 @@ main(int argc, char **argv) {
 			opts = 1;
 			break;
 		case 'u':
-			display_in_cyl_units = 0;
+			display_in_cyl_units = 0;		/* default */
+			if (optarg && strcmp(optarg, "=cylinders") == 0)
+				display_in_cyl_units = !display_in_cyl_units;
+			else if (optarg && strcmp(optarg, "=sectors"))
+				fatal(usage);
 			break;
 		case 'V':
 		case 'v':
