@@ -16,12 +16,8 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#ifdef HAVE_TERMIOS_H
 #include <termios.h>
-#endif
-#ifdef HAVE_TERMIO_H
-#include <termio.h>
-#endif
+#include <errno.h>
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
@@ -33,14 +29,19 @@ extern char *optarg;
 extern int optind;
 #endif
 
-#define OUTPUT_VALUE_ONLY	0x0001
-#define OUTPUT_DEVICE_ONLY	0x0002
-#define OUTPUT_PRETTY_LIST	0x0004
-#define OUTPUT_UDEV_LIST	0x0008
+#define OUTPUT_VALUE_ONLY	(1 << 1)
+#define OUTPUT_DEVICE_ONLY	(1 << 2)
+#define OUTPUT_PRETTY_LIST	(1 << 3)
+#define OUTPUT_UDEV_LIST	(1 << 4)
+#define OUTPUT_EXPORT_LIST	(1 << 5)
+
+#define LOWPROBE_TOPOLOGY	(1 << 1)
+#define LOWPROBE_SUPERBLOCKS	(1 << 2)
 
 #include <blkid.h>
 
 #include "ismounted.h"
+#include "strtosize.h"
 
 const char *progname = "blkid";
 
@@ -60,13 +61,15 @@ static void usage(int error)
 		"  %1$s -L <label> | -U <uuid>\n\n"
 		"  %1$s [-c <file>] [-ghlLv] [-o format] [-s <tag>] \n"
 		"        [-t <token>] [-w <file>] [dev ...]\n\n"
-		"  %1$s -p [-O <offset>] [-S <size>] [-o format] <dev> [dev ...]\n\n"
+		"  %1$s -p [-s <tag>] [-O <offset>] [-S <size>] \n"
+		"        [-o format] <dev> [dev ...]\n\n"
+		"  %1$s -i [-s <tag>] [-o format] <dev> [dev ...]\n\n"
 		"Options:\n"
 		"  -c <file>   cache file (default: /etc/blkid.tab, /dev/null = none)\n"
 		"  -h          print this usage message and exit\n"
 		"  -g          garbage collect the blkid cache\n"
 		"  -o <format> output format; can be one of:\n"
-		"              value, device, list, udev or full; (default: full)\n"
+		"              value, device, list, udev, export or full; (default: full)\n"
 		"  -s <tag>    show specified tag(s) (default show all tags)\n"
 		"  -t <token>  find device with a specific token (NAME=value pair)\n"
 		"  -l          lookup the the first device with arguments specified by -t\n"
@@ -76,10 +79,12 @@ static void usage(int error)
 		"  -w <file>   write cache to different file (/dev/null = no write)\n"
 		"  <dev>       specify device(s) to probe (default: all devices)\n\n"
 		"Low-level probing options:\n"
-		"  -p          switch to low-level mode (bypass cache)\n"
-		"  -S <bytes>  overwrite device size\n"
-		"  -O <bytes>  probe at the given offset\n"
+		"  -p          low-level superblocks probing (bypass cache)\n"
+		"  -i          gather information about I/O limits\n"
+		"  -S <size>   overwrite device size\n"
+		"  -O <offset> probe at the given offset\n"
 		"  -u <list>   filter by \"usage\" (e.g. -u filesystem,raid)\n"
+		"  -n <list>   filter by filesystem type (e.g. -n vfat,ext3)\n"
 		"\n",
 				progname);
 
@@ -239,6 +244,7 @@ static void pretty_print_dev(blkid_dev dev)
 static void print_udev_format(const char *name, const char *value, size_t sz)
 {
 	char enc[265], safe[256];
+	size_t namelen = strlen(name);
 
 	*safe = *enc = '\0';
 
@@ -255,7 +261,24 @@ static void print_udev_format(const char *name, const char *value, size_t sz)
 
 		blkid_encode_string(value, enc, sizeof(enc));
 		printf("ID_FS_%s_ENC=%s\n", name, enc);
-	}
+
+	} else if (!strcmp(name, "PTTYPE")) {
+		printf("ID_PART_TABLE_TYPE=%s\n", value);
+
+	} else if (!strcmp(name, "PART_ENTRY_NAME") ||
+		  !strcmp(name, "PART_ENTRY_TYPE")) {
+
+		blkid_encode_string(value, enc, sizeof(enc));
+		printf("ID_%s=%s\n", name, enc);
+
+	} else if (!strncmp(name, "PART_ENTRY_", 11))
+		printf("ID_%s=%s\n", name, value);
+
+	else if (namelen >= 15 && (
+		   !strcmp(name + (namelen - 12), "_SECTOR_SIZE") ||
+		   !strcmp(name + (namelen - 8), "_IO_SIZE") ||
+		   !strcmp(name, "ALIGNMENT_OFFSET")))
+			printf("ID_IOLIMIT_%s=%s\n", name, value);
 	else
 		printf("ID_FS_%s=%s\n", name, value);
 }
@@ -280,6 +303,12 @@ static void print_value(int output, int num, const char *devname,
 	} else if (output & OUTPUT_UDEV_LIST) {
 		print_udev_format(name, value, valsz);
 
+	} else if (output & OUTPUT_EXPORT_LIST) {
+		fputs(name, stdout);
+		fputs("=", stdout);
+		safe_print(value, valsz);
+		fputs("\n", stdout);
+
 	} else {
 		if (num == 1 && devname)
 			printf("%s: ", devname);
@@ -295,6 +324,7 @@ static void print_tags(blkid_dev dev, char *show[], int output)
 	blkid_tag_iterate	iter;
 	const char		*type, *value, *devname;
 	int			num = 1;
+	static int		first = 1;
 
 	if (!dev)
 		return;
@@ -315,12 +345,22 @@ static void print_tags(blkid_dev dev, char *show[], int output)
 	while (blkid_tag_next(iter, &type, &value) == 0) {
 		if (show[0] && !has_item(show, type))
 			continue;
+
+		if (num == 1 && !first &&
+		    (output & (OUTPUT_UDEV_LIST | OUTPUT_EXPORT_LIST)))
+			/* add extra line between output from more devices */
+			fputc('\n', stdout);
+
 		print_value(output, num++, devname, value, type, strlen(value));
 	}
 	blkid_tag_iterate_end(iter);
 
-	if (num > 1 && !(output & (OUTPUT_VALUE_ONLY | OUTPUT_UDEV_LIST)))
-		printf("\n");
+	if (num > 1) {
+		if (!(output & (OUTPUT_VALUE_ONLY | OUTPUT_UDEV_LIST |
+						OUTPUT_EXPORT_LIST)))
+			printf("\n");
+		first = 0;
+	}
 }
 
 
@@ -401,8 +441,54 @@ done:
 	return rc;
 }
 
-static int lowprobe_device(blkid_probe pr, const char *devname,	char *show[],
-			int output, blkid_loff_t offset, blkid_loff_t size)
+static int lowprobe_superblocks(blkid_probe pr)
+{
+	struct stat st;
+	int rc;
+
+	if (fstat(blkid_probe_get_fd(pr), &st))
+		return -1;
+
+	blkid_probe_enable_partitions(pr, 1);
+	blkid_probe_set_partitions_flags(pr, BLKID_PARTS_ENTRY_DETAILS);
+
+	if (!S_ISCHR(st.st_mode) && blkid_probe_get_size(pr) <= 1024 * 1440) {
+		/*
+		 * check if the small disk is partitioned, if yes then
+		 * don't probe for filesystems.
+		 */
+		blkid_probe_enable_superblocks(pr, 0);
+
+		rc = blkid_do_fullprobe(pr);
+		if (rc < 0)
+			return rc;	/* -1 = error, 1 = nothing, 0 = succes */
+
+		if (blkid_probe_lookup_value(pr, "PTTYPE", NULL, NULL) == 0)
+			return 0;	/* partition table detected */
+
+		/* small whole-disk is unpartitioned, probe for filesystems only */
+		blkid_probe_enable_partitions(pr, 0);
+	}
+
+	blkid_probe_enable_superblocks(pr, 1);
+
+	return blkid_do_safeprobe(pr);
+}
+
+static int lowprobe_topology(blkid_probe pr)
+{
+	/* enable topology probing only */
+	blkid_probe_enable_topology(pr, 1);
+
+	blkid_probe_enable_superblocks(pr, 0);
+	blkid_probe_enable_partitions(pr, 0);
+
+	return blkid_do_fullprobe(pr);
+}
+
+static int lowprobe_device(blkid_probe pr, const char *devname,
+			int chain, char *show[], int output,
+			blkid_loff_t offset, blkid_loff_t size)
 {
 	const char *data;
 	const char *name;
@@ -410,19 +496,29 @@ static int lowprobe_device(blkid_probe pr, const char *devname,	char *show[],
 	size_t len;
 	int fd;
 	int rc = 0;
+	static int first = 1;
 
 	fd = open(devname, O_RDONLY);
-	if (fd < 0)
+	if (fd < 0) {
+		fprintf(stderr, "error: %s: %s\n", devname, strerror(errno));
 		return 2;
-
+	}
 	if (blkid_probe_set_device(pr, fd, offset, size))
 		goto done;
-	rc = blkid_do_safeprobe(pr);
-	if (rc)
+
+	if (chain & LOWPROBE_TOPOLOGY)
+		rc = lowprobe_topology(pr);
+	if (rc >= 0 && (chain & LOWPROBE_SUPERBLOCKS))
+		rc = lowprobe_superblocks(pr);
+	if (rc < 0)
 		goto done;
 
 	nvals = blkid_probe_numof_values(pr);
 
+	if (nvals && !first && output & (OUTPUT_UDEV_LIST | OUTPUT_EXPORT_LIST))
+		/* add extra line between output from devices */
+		fputc('\n', stdout);
+
 	if (output & OUTPUT_DEVICE_ONLY) {
 		printf("%s\n", devname);
 		goto done;
@@ -437,7 +533,10 @@ static int lowprobe_device(blkid_probe pr, const char *devname,	char *show[],
 		print_value(output, num++, devname, (char *) data, name, len);
 	}
 
-	if (nvals > 1 && !(output & (OUTPUT_VALUE_ONLY | OUTPUT_UDEV_LIST)))
+	if (first)
+		first = 0;
+	if (nvals >= 1 && !(output & (OUTPUT_VALUE_ONLY |
+					OUTPUT_UDEV_LIST | OUTPUT_EXPORT_LIST)))
 		printf("\n");
 done:
 	if (rc == -2) {
@@ -458,32 +557,89 @@ done:
 static int list_to_usage(const char *list, int *flag)
 {
 	int mask = 0;
-	const char *word, *p = list;
+	const char *word = NULL, *p = list;
 
 	if (p && strncmp(p, "no", 2) == 0) {
 		*flag = BLKID_FLTR_NOTIN;
 		p += 2;
 	}
-
-	for (word = p; p && *p; p++) {
-		if (*p == ',' || *(p + 1) == '\0') {
-			if (!strncmp(word, "filesystem", 10))
-				mask |= BLKID_USAGE_FILESYSTEM;
-			else if (!strncmp(word, "raid", 4))
-				mask |= BLKID_USAGE_RAID;
-			else if (!strncmp(word, "crypto", 6))
-				mask |= BLKID_USAGE_CRYPTO;
-			else if (!strncmp(word, "other", 5))
-				mask |= BLKID_USAGE_OTHER;
-			else {
-				fprintf(stderr, "unknown usage keyword '%*s'\n",
-						(int) (p - word), word);
-				exit(4);
-			}
-			word = p + 1;
-		}
+	if (!p || !*p)
+		goto err;
+	while(p) {
+		word = p;
+		p = strchr(p, ',');
+		if (p)
+			p++;
+		if (!strncmp(word, "filesystem", 10))
+			mask |= BLKID_USAGE_FILESYSTEM;
+		else if (!strncmp(word, "raid", 4))
+			mask |= BLKID_USAGE_RAID;
+		else if (!strncmp(word, "crypto", 6))
+			mask |= BLKID_USAGE_CRYPTO;
+		else if (!strncmp(word, "other", 5))
+			mask |= BLKID_USAGE_OTHER;
+		else
+			goto err;
 	}
 	return mask;
+err:
+	*flag = 0;
+	fprintf(stderr, "unknown kerword in -u <list> argument: '%s'\n",
+			word ? word : list);
+	exit(4);
+}
+
+/* converts comma separated list to types[] */
+static char **list_to_types(const char *list, int *flag)
+{
+	int i;
+	const char *p = list;
+	char **res;
+
+	if (p && strncmp(p, "no", 2) == 0) {
+		*flag = BLKID_FLTR_NOTIN;
+		p += 2;
+	}
+	if (!p || !*p) {
+		fprintf(stderr, "error: -u <list> argument is empty\n");
+		goto err;
+	}
+	for (i = 1; p && (p = strchr(p, ',')); i++, p++);
+
+	res = calloc(i + 1, sizeof(char *));
+	if (!res)
+		goto err_mem;
+	p = *flag & BLKID_FLTR_NOTIN ? list + 2 : list;
+	i = 0;
+
+	while(p) {
+		const char *word = p;
+		p = strchr(p, ',');
+		res[i] = p ? strndup(word, p - word) : strdup(word);
+		if (!res[i++])
+			goto err_mem;
+		if (p)
+			p++;
+	}
+	res[i] = NULL;
+	return res;
+err_mem:
+	fprintf(stderr, "out of memory\n");
+err:
+	*flag = 0;
+	free(res);
+	exit(4);
+}
+
+static void free_types_list(char *list[])
+{
+	char **n;
+
+	if (!list)
+		return;
+	for (n = list; *n; n++)
+		free(*n);
+	free(list);
 }
 
 int main(int argc, char **argv)
@@ -495,6 +651,7 @@ int main(int argc, char **argv)
 	char *read = NULL;
 	char *write = NULL;
 	int fltr_usage = 0;
+	char **fltr_type = NULL;
 	int fltr_flag = BLKID_FLTR_ONLYIN;
 	unsigned int numdev = 0, numtag = 0;
 	int version = 0;
@@ -503,11 +660,11 @@ int main(int argc, char **argv)
 	int output_format = 0;
 	int lookup = 0, gc = 0, lowprobe = 0, eval = 0;
 	int c;
-	blkid_loff_t offset = 0, size = 0;
+	uintmax_t offset = 0, size = 0;
 
 	show[0] = NULL;
 
-	while ((c = getopt (argc, argv, "c:f:ghlL:o:O:ps:S:t:u:U:w:v")) != EOF)
+	while ((c = getopt (argc, argv, "c:f:ghilL:n:o:O:ps:S:t:u:U:w:v")) != EOF)
 		switch (c) {
 		case 'c':
 			if (optarg && !*optarg)
@@ -522,7 +679,18 @@ int main(int argc, char **argv)
 			search_value = strdup(optarg);
 			search_type = strdup("LABEL");
 			break;
+		case 'n':
+			if (fltr_usage) {
+				fprintf(stderr, "error: -u and -n options are mutually exclusive\n");
+				exit(4);
+			}
+			fltr_type = list_to_types(optarg, &fltr_flag);
+			break;
 		case 'u':
+			if (fltr_type) {
+				fprintf(stderr, "error: -u and -n options are mutually exclusive\n");
+				exit(4);
+			}
 			fltr_usage = list_to_usage(optarg, &fltr_flag);
 			break;
 		case 'U':
@@ -530,6 +698,9 @@ int main(int argc, char **argv)
 			search_value = strdup(optarg);
 			search_type = strdup("UUID");
 			break;
+		case 'i':
+			lowprobe |= LOWPROBE_TOPOLOGY;
+			break;
 		case 'l':
 			lookup++;
 			break;
@@ -545,6 +716,8 @@ int main(int argc, char **argv)
 				output_format = OUTPUT_PRETTY_LIST;
 			else if (!strcmp(optarg, "udev"))
 				output_format = OUTPUT_UDEV_LIST;
+			else if (!strcmp(optarg, "export"))
+				output_format = OUTPUT_EXPORT_LIST;
 			else if (!strcmp(optarg, "full"))
 				output_format = 0;
 			else {
@@ -555,10 +728,13 @@ int main(int argc, char **argv)
 			}
 			break;
 		case 'O':
-			offset = strtoll(optarg, NULL, 10);
+			if (strtosize(optarg, &offset))
+				fprintf(stderr,
+					"Invalid offset '%s' specified\n",
+					optarg);
 			break;
 		case 'p':
-			lowprobe++;
+			lowprobe |= LOWPROBE_SUPERBLOCKS;
 			break;
 		case 's':
 			if (numtag + 1 >= sizeof(show) / sizeof(*show)) {
@@ -569,7 +745,10 @@ int main(int argc, char **argv)
 			show[numtag] = NULL;
 			break;
 		case 'S':
-			size = strtoll(optarg, NULL, 10);
+			if (strtosize(optarg, &size))
+				fprintf(stderr,
+					"Invalid size '%s' specified\n",
+					optarg);
 			break;
 		case 't':
 			if (search_type) {
@@ -624,7 +803,7 @@ int main(int argc, char **argv)
 	}
 	err = 2;
 
-	if (eval == 0 && output_format & OUTPUT_PRETTY_LIST) {
+	if (eval == 0 && (output_format & OUTPUT_PRETTY_LIST)) {
 		if (lowprobe) {
 			fprintf(stderr, "The low-level probing mode does not "
 					"support 'list' output format\n");
@@ -644,24 +823,35 @@ int main(int argc, char **argv)
 					"requires a device\n");
 			exit(4);
 		}
+
+		/* automatically enable 'export' format for I/O Limits */
+		if (!output_format  && (lowprobe & LOWPROBE_TOPOLOGY))
+			output_format = OUTPUT_EXPORT_LIST;
+
 		pr = blkid_new_probe();
 		if (!pr)
 			goto exit;
 
-		blkid_probe_enable_superblocks(pr, 1);
-
-		blkid_probe_set_superblocks_flags(pr,
+		if (lowprobe & LOWPROBE_SUPERBLOCKS) {
+			blkid_probe_set_superblocks_flags(pr,
 				BLKID_SUBLKS_LABEL | BLKID_SUBLKS_UUID |
 				BLKID_SUBLKS_TYPE | BLKID_SUBLKS_SECTYPE |
 				BLKID_SUBLKS_USAGE | BLKID_SUBLKS_VERSION);
 
-		if (fltr_usage &&
-		    blkid_probe_filter_superblocks_usage(pr, fltr_flag, fltr_usage))
-			goto exit;
+			if (fltr_usage && blkid_probe_filter_superblocks_usage(
+						pr, fltr_flag, fltr_usage))
+				goto exit;
+
+			else if (fltr_type && blkid_probe_filter_superblocks_type(
+						pr, fltr_flag, fltr_type))
+				goto exit;
+		}
 
 		for (i = 0; i < numdev; i++)
-			err = lowprobe_device(pr, devices[i], show,
-					output_format, offset, size);
+			err = lowprobe_device(pr, devices[i], lowprobe, show,
+					output_format,
+					(blkid_loff_t) offset,
+					(blkid_loff_t) size);
 		blkid_free_probe(pr);
 	} else if (eval) {
 		/*
@@ -727,6 +917,7 @@ int main(int argc, char **argv)
 exit:
 	free(search_type);
 	free(search_value);
+	free_types_list(fltr_type);
 	if (!lowprobe && !eval)
 		blkid_put_cache(cache);
 	return err;
