@@ -41,6 +41,7 @@
 #include "env.h"
 #include "nls.h"
 #include "blkdev.h"
+#include "strtosize.h"
 
 #define DO_PS_FIDDLING
 
@@ -313,11 +314,11 @@ append_opt(char *s, const char *opt, const char *val)
 }
 
 static char *
-append_numopt(char *s, const char *opt, long num)
+append_numopt(char *s, const char *opt, unsigned int num)
 {
 	char buf[32];
 
-	snprintf(buf, sizeof(buf), "%ld", num);
+	snprintf(buf, sizeof(buf), "%u", num);
 	return append_opt(s, opt, buf);
 }
 
@@ -648,74 +649,89 @@ do_mount (struct mountargs *args, int *special, int *status) {
 static int
 check_special_mountprog(const char *spec, const char *node, const char *type, int flags,
 			char *extra_opts, int *status) {
-  char mountprog[120];
-  struct stat statbuf;
-  int res;
+	char search_path[] = FS_SEARCH_PATH;
+	char *path, mountprog[150];
+	struct stat statbuf;
+	int res;
 
-  if (!external_allowed)
-      return 0;
+	if (!external_allowed)
+		return 0;
 
-  if (type == NULL || strcmp(type, "none") == 0)
-	  return 0;
+	if (type == NULL || strcmp(type, "none") == 0)
+		return 0;
 
-  if (strlen(type) < 100) {
-       sprintf(mountprog, "/sbin/mount.%s", type);
-       if (stat(mountprog, &statbuf) == 0) {
-	    if (verbose)
-		 fflush(stdout);
-	    res = fork();
-	    if (res == 0) {
-		 char *oo, *mountargs[10];
-		 int i = 0;
+	path = strtok(search_path, ":");
+	while (path) {
+		res = snprintf(mountprog, sizeof(mountprog), "%s/mount.%s",
+			       path, type);
+		path = strtok(NULL, ":");
+		if (res >= sizeof(mountprog) || res < 0)
+			continue;
 
-		 if(setgid(getgid()) < 0)
-			 die(EX_FAIL, _("mount: cannot set group id: %s"), strerror(errno));
+		if (stat(mountprog, &statbuf))
+			continue;
 
-		 if(setuid(getuid()) < 0)
-			 die(EX_FAIL, _("mount: cannot set user id: %s"), strerror(errno));
-
-		 oo = fix_opts_string (flags, extra_opts, NULL);
-		 mountargs[i++] = mountprog;				/* 1 */
-		 mountargs[i++] = (char *) spec;			/* 2 */
-		 mountargs[i++] = (char *) node;			/* 3 */
-		 if (sloppy && strncmp(type, "nfs", 3) == 0)
-		      mountargs[i++] = "-s";				/* 4 */
-		 if (fake)
-		      mountargs[i++] = "-f";				/* 5 */
-		 if (nomtab)
-		      mountargs[i++] = "-n";				/* 6 */
-		 if (verbose)
-		      mountargs[i++] = "-v";				/* 7 */
-		 if (oo && *oo) {
-		      mountargs[i++] = "-o";				/* 8 */
-		      mountargs[i++] = oo;				/* 9 */
-		 }
-		 mountargs[i] = NULL;					/* 10 */
-
-		 if (verbose > 2) {
-			i = 0;
-			while(mountargs[i]) {
-				printf("mount: external mount: argv[%d] = \"%s\"\n",
-					i, mountargs[i]);
-				i++;
-			}
+		if (verbose)
 			fflush(stdout);
-		 }
 
-		 execv(mountprog, mountargs);
-		 exit(1);	/* exec failed */
-	    } else if (res != -1) {
-		 int st;
-		 wait(&st);
-		 *status = (WIFEXITED(st) ? WEXITSTATUS(st) : EX_SYSERR);
-		 return 1;
-	    } else {
-		 int errsv = errno;
-		 error(_("mount: cannot fork: %s"), strerror(errsv));
-	    }
-       }
-  }
-  return 0;
+		switch (fork()) {
+		case 0: { /* child */
+			char *oo, *mountargs[10];
+			int i = 0;
+
+			if (setgid(getgid()) < 0)
+				die(EX_FAIL, _("mount: cannot set group id: %s"), strerror(errno));
+
+			if (setuid(getuid()) < 0)
+				die(EX_FAIL, _("mount: cannot set user id: %s"), strerror(errno));
+
+			oo = fix_opts_string (flags, extra_opts, NULL);
+			mountargs[i++] = mountprog;			/* 1 */
+			mountargs[i++] = (char *) spec;			/* 2 */
+			mountargs[i++] = (char *) node;			/* 3 */
+			if (sloppy && strncmp(type, "nfs", 3) == 0)
+				mountargs[i++] = "-s";			/* 4 */
+			if (fake)
+				mountargs[i++] = "-f";			/* 5 */
+			if (nomtab)
+				mountargs[i++] = "-n";			/* 6 */
+			if (verbose)
+				mountargs[i++] = "-v";			/* 7 */
+			if (oo && *oo) {
+				mountargs[i++] = "-o";			/* 8 */
+				mountargs[i++] = oo;			/* 9 */
+			}
+			mountargs[i] = NULL;				/* 10 */
+
+			if (verbose > 2) {
+				i = 0;
+				while (mountargs[i]) {
+					printf("mount: external mount: argv[%d] = \"%s\"\n",
+						i, mountargs[i]);
+					i++;
+				}
+				fflush(stdout);
+			}
+
+			execv(mountprog, mountargs);
+			exit(1);	/* exec failed */
+		}
+
+		default: { /* parent */
+			int st;
+			wait(&st);
+			*status = (WIFEXITED(st) ? WEXITSTATUS(st) : EX_SYSERR);
+			return 1;
+		}
+
+		case -1: { /* error */
+			int errsv = errno;
+			error(_("mount: cannot fork: %s"), strerror(errsv));
+		}
+		}
+	}
+
+	return 0;
 }
 
 
@@ -729,8 +745,6 @@ static int
 was_tested(const char *fstype) {
 	struct tried *t;
 
-	if (fsprobe_known_fstype(fstype))
-		return 1;
 	for (t = tried; t; t = t->next) {
 		if (!strcmp(t->type, fstype))
 			return 1;
@@ -767,8 +781,8 @@ procfsnext(FILE *procfs) {
    char fsname[100];
 
    while (fgets(line, sizeof(line), procfs)) {
-      if (sscanf (line, "nodev %[^\n]\n", fsname) == 1) continue;
-      if (sscanf (line, " %[^ \n]\n", fsname) != 1) continue;
+      if (sscanf (line, "nodev %[^#\n]\n", fsname) == 1) continue;
+      if (sscanf (line, " %[^# \n]\n", fsname) != 1) continue;
       return xstrdup(fsname);
    }
    return 0;
@@ -871,9 +885,9 @@ procfsloop_mount(int (*mount_fn)(struct mountargs *, int *, int *),
 }
 
 static const char *
-guess_fstype_by_devname(const char *devname)
+guess_fstype_by_devname(const char *devname, int *ambivalent)
 {
-   const char *type = fsprobe_get_fstype_by_devname(devname);
+   const char *type = fsprobe_get_fstype_by_devname_ambi(devname, ambivalent);
 
    if (verbose) {
       printf (_("mount: you didn't specify a filesystem type for %s\n"), devname);
@@ -900,12 +914,13 @@ static int
 guess_fstype_and_mount(const char *spec, const char *node, const char **types,
 		       int flags, char *mount_opts, int *special, int *status) {
    struct mountargs args = { spec, node, NULL, flags & ~MS_NOSYS, mount_opts };
+   int ambivalent = 0;
 
    if (*types && strcasecmp (*types, "auto") == 0)
       *types = NULL;
 
    if (!*types && !(flags & MS_REMOUNT)) {
-      *types = guess_fstype_by_devname(spec);
+      *types = guess_fstype_by_devname(spec, &ambivalent);
       if (*types) {
 	  if (!strcmp(*types, MNTTYPE_SWAP)) {
 	      error(_("%s looks like swapspace - not mounted"), spec);
@@ -915,6 +930,11 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
 	      args.type = *types;
 	      return do_mount (&args, special, status);
           }
+      } else if (ambivalent) {
+          error(_("mount: %s: more filesystems detected. This should not happen,\n"
+		  "       use -t <type> to explicitly specify the filesystem type or\n"
+		  "       use wipefs(8) to clean up the device.\n"), spec);
+	  return 1;
       }
    }
 
@@ -1051,12 +1071,27 @@ is_mounted_same_loopfile(const char *node0, const char *loopfile, unsigned long
 	return res;
 }
 
+static int
+parse_offset(const char **opt, uintmax_t *val)
+{
+	char *tmp;
+
+	if (strtosize(*opt, val))
+		return -1;
+
+	tmp = xmalloc(32);
+	snprintf(tmp, 32, "%jd", *val);
+	my_free(*opt);
+	*opt = tmp;
+	return 0;
+}
+
 static int
 loop_check(const char **spec, const char **type, int *flags,
 	   int *loop, const char **loopdev, const char **loopfile,
 	   const char *node) {
   int looptype;
-  unsigned long long offset, sizelimit;
+  uintmax_t offset = 0, sizelimit = 0;
 
   /*
    * In the case of a loop mount, either type is of the form lo@/dev/loop5
@@ -1084,6 +1119,22 @@ loop_check(const char **spec, const char **type, int *flags,
   *loop = ((*flags & MS_LOOP) || *loopdev || opt_offset || opt_sizelimit || opt_encryption);
   *loopfile = *spec;
 
+  /* Automatically create a loop device from a regular file if a filesystem
+   * is not specified or the filesystem is known for libblkid (these
+   * filesystems work with block devices only).
+   *
+   * Note that there is not a restriction (on kernel side) that prevents regular
+   * file as a mount(2) source argument. A filesystem that is able to mount
+   * regular files could be implemented.
+   */
+  if (!*loop && !(*flags & (MS_BIND | MS_MOVE | MS_PROPAGATION)) &&
+      (!*type || strcmp(*type, "auto") == 0 || fsprobe_known_fstype(*type))) {
+
+    struct stat st;
+    if (stat(*loopfile, &st) == 0)
+      *loop = S_ISREG(st.st_mode);
+  }
+
   if (*loop) {
     *flags |= MS_LOOP;
     if (fake) {
@@ -1096,8 +1147,14 @@ loop_check(const char **spec, const char **type, int *flags,
       if (*flags & MS_RDONLY)
         loop_opts |= SETLOOP_RDONLY;
 
-      offset = opt_offset ? strtoull(opt_offset, NULL, 0) : 0;
-      sizelimit = opt_sizelimit ? strtoull(opt_sizelimit, NULL, 0) : 0;
+      if (opt_offset && parse_offset(&opt_offset, &offset)) {
+        error(_("mount: invalid offset '%s' specified"), opt_offset);
+        return EX_FAIL;
+      }
+      if (opt_sizelimit && parse_offset(&opt_sizelimit, &sizelimit)) {
+        error(_("mount: invalid sizelimit '%s' specified"), opt_sizelimit);
+        return EX_FAIL;
+      }
 
       if (is_mounted_same_loopfile(node, *loopfile, offset)) {
         error(_("mount: according to mtab %s is already mounted on %s as loop"), *loopfile, node);
@@ -1290,7 +1347,6 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   int loop = 0;
   const char *loopdev = 0, *loopfile = 0;
   struct stat statbuf;
-  int retries = 0;	/* Nr of retries for mount in case of ENOMEDIUM */
 
   /* copies for freeing on exit */
   const char *opts1, *spec1, *node1, *types1, *extra_opts1;
@@ -1356,7 +1412,6 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       goto out;
   }
 
-mount_retry:
   block_signals (SIG_BLOCK);
 
   if (!fake) {
@@ -1371,7 +1426,7 @@ mount_retry:
   }
 
   /* Kernel allows to use MS_RDONLY for bind mounts, but the read-only request
-   * could be silently ignored. Check it to avoid 'ro' in ntab and 'rw' in
+   * could be silently ignored. Check it to avoid 'ro' in mtab and 'rw' in
    * /proc/mounts.
    */
   if (!fake && mnt5_res == 0 &&
@@ -1381,17 +1436,31 @@ mount_retry:
       flags &= ~MS_RDONLY;
   }
 
+  /* Kernel can silently add MS_RDONLY flag when mounting file system that
+   * does not have write support. Check this to avoid 'ro' in /proc/mounts
+   * and 'rw' in mtab.
+   */
+  if (!fake && mnt5_res == 0 &&
+      !(flags & (MS_RDONLY | MS_PROPAGATION | MS_MOVE)) &&
+      is_readonly(node)) {
+
+      printf(_("mount: warning: %s seems to be mounted read-only.\n"), node);
+      flags |= MS_RDONLY;
+  }
+
   if (fake || mnt5_res == 0) {
       /* Mount succeeded, report this (if verbose) and write mtab entry.  */
 
       if (!(mounttype & MS_PROPAGATION)) {
+	      char *mtab_opts = fix_opts_string (flags & ~MS_NOMTAB, extra_opts, user);
 	      update_mtab_entry(loop ? loopfile : spec,
 			node,
 			types ? types : "unknown",
-			fix_opts_string (flags & ~MS_NOMTAB, extra_opts, user),
+			mtab_opts,
 			flags,
 			freq,
 			pass);
+	      free (mtab_opts);
       }
 
       block_signals (SIG_UNBLOCK);
@@ -1602,14 +1671,6 @@ mount_retry:
       break;
     }
     case ENOMEDIUM:
-      if (retries < CRDOM_NOMEDIUM_RETRIES) {
-	      if (verbose)
-		      printf(_("mount: no medium found on %s ...trying again\n"),
-				 spec);
-              sleep(3);
-	      ++retries;
-              goto mount_retry;
-      }
       error(_("mount: no medium found on %s"), spec);
       break;
     default:
@@ -1670,12 +1731,12 @@ usersubst(const char *opts) {
 
 	s = "uid=useruid";
 	if (opts && (w = strstr(opts, s)) != NULL) {
-		sprintf(id, "uid=%d", getuid());
+		sprintf(id, "uid=%u", getuid());
 		opts = subst_string(opts, w, strlen(s), id);
 	}
 	s = "gid=usergid";
 	if (opts && (w = strstr(opts, s)) != NULL) {
-		sprintf(id, "gid=%d", getgid());
+		sprintf(id, "gid=%u", getgid());
 		opts = subst_string(opts, w, strlen(s), id);
 	}
 	return xstrdup(opts);
@@ -2240,8 +2301,8 @@ main(int argc, char *argv[]) {
 		printf("mount: mtab path:  \"%s\"\n", _PATH_MOUNTED);
 		printf("mount: lock path:  \"%s\"\n", _PATH_MOUNTED_LOCK);
 		printf("mount: temp path:  \"%s\"\n", _PATH_MOUNTED_TMP);
-		printf("mount: UID:        %d\n", getuid());
-		printf("mount: eUID:       %d\n", geteuid());
+		printf("mount: UID:        %u\n", getuid());
+		printf("mount: eUID:       %u\n", geteuid());
 	}
 
 	argc -= optind;
@@ -2272,7 +2333,7 @@ main(int argc, char *argv[]) {
 			if (ruid == 0 && euid != 0)
 				/* user is root, but setuid to non-root */
 				die (EX_USAGE, _("mount: only root can do that "
-					"(effective UID is %d)"), euid);
+					"(effective UID is %u)"), euid);
 
 			die (EX_USAGE, _("mount: only root can do that"));
 		}
